﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace FileMapped
{
    unsafe public class FileMm : IDisposable
    {
        private const uint PAGE_READONLY = 0x00000002;
        private const uint PAGE_READWRITE = 0x00000004;
        private const uint FILE_MAP_READ = 0x00000004;
        private const uint FILE_MAP_ALL_ACCESS = 0x00000002;
        private const Int64 TWO_TO_32 = 0x100000000;

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern SafeFileHandle CreateFile(String lpFileName,
            FileAccess dwDesiredAccess, FileShare dwShareMode,
            IntPtr lpSecurityAttributes, FileMode dwCreationDisposition,
            uint dwFlagsAndAttributes, IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool CloseHandle(SafeFileHandle h);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool SetFilePointerEx(SafeFileHandle h, Int64 Distance, Int64* pPosition, uint dwMoveMethod);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern UInt32 SetFilePointer(SafeFileHandle h, UInt32 DistanceLow, Int32* pDistanceHigh, uint dwMoveMethod);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool SetEndOfFile(SafeFileHandle h);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern SafeFileHandle CreateFileMapping(SafeFileHandle hFile,
            IntPtr lpSecurityAttributes, UInt32 dwProtect, UInt32 dwMaximumSizeHigh,
            UInt32 dwMaximumSizeLow, IntPtr lpMapName);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern byte* MapViewOfFile(SafeFileHandle hMapObject,
            UInt32 dwAccess, Int32 dwOffsetHigh, UInt32 dwOffsetLow, UInt64 dwNumberOfBytesToMap);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern bool UnmapViewOfFile(byte* lpBaseAddress);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern int GetLastError();

        /// <summary>
        /// Mmeory Mapped file access common constructor.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="readOnly">Determines mode for file mapping</param>
        /// <param name="fileSize">If >0 and !readOnly, set the file to this length</param>
        private void commonConstructor (String pathName, bool readOnly, Int64 fileSize)
        {
            inHandle = CreateFile(pathName, readOnly ? FileAccess.Read : FileAccess.ReadWrite,
                                            readOnly ? FileShare.Read : FileShare.ReadWrite, IntPtr.Zero, 
                                            readOnly ? FileMode.Open : FileMode.Create, 0, IntPtr.Zero);
            if (inHandle.IsInvalid) { throw new Exception("Error. Failed to open file " + pathName); }

            _fileSize = fileSize;

            if (!readOnly && fileSize > 0)
            {
                // Set the file size if it's read-write
                Int32 distanceHigh = (Int32)(_fileSize / TWO_TO_32);
                UInt32 distanceLow = (UInt32)(_fileSize % TWO_TO_32);
                if ( distanceLow != SetFilePointer(inHandle, distanceLow, &distanceHigh, 0) )
                    { throw new Exception("Failed to SetFilePointer for read-write file"); }
                if (!SetEndOfFile(inHandle)) { throw new Exception("Failed to SetEndOfFile for read-write file"); }
            }
            SafeFileHandle inMapHandle = CreateFileMapping(inHandle, IntPtr.Zero, readOnly ? PAGE_READONLY : PAGE_READWRITE, 0, 0, IntPtr.Zero);
            if (inMapHandle == null) { throw new Exception("Failed to CreateFileMapping for mapping"); }
            _pFile = (byte*)MapViewOfFile(inMapHandle, readOnly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS, 0, 0, (UInt64)0); CloseHandle(inMapHandle);
            if (_pFile == null) { throw new Exception("Error. Failed to map view of file"); }
        }

        public FileMm(String fileName, bool readOnly, Int64 fileSize)
        {
            commonConstructor(fileName, readOnly, fileSize);
        }
        public FileMm(String fileName)
        {
            commonConstructor(fileName, true, 0);
        }

        private Int64 _fileSize;
        private byte* _pFile = null;
        private SafeFileHandle inHandle = null;

        public Int64 FileSize { get { return _fileSize; } }
        public byte* pFile { get { return _pFile; } }

        public void Dispose()
        {
            if (pFile != null) UnmapViewOfFile(pFile);
            if (!inHandle.IsInvalid) CloseHandle(inHandle);
        }
    }
}
